/*   AUTHOR : Tomer Moscovich (tomer@moscovich.net)
 *   Copyright (c) INRIA, 2009. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 *
 * $Id$
 */

package net.claribole.zgrviewer;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

import fr.inria.zvtm.glyphs.*;

/**
 * Given a GeneralPath and a mouse cursor position, calculate path cursor position and scale
 * 
 * Usage:
 *    lsc = new LinkSliderCalc(edge.getPathReference(), myViewportWidth));
 *    lsc.updateMousePosition(newMousePos);
 *    patPoint = lsc.getPositionAlongPath();
 *    scale = lsc.getScale();
 */

public class LinkSliderCalc {

    public PolyLine polyLine;

    protected Point2D mousePos;
    protected Point2D nearestPoint;
    protected Line2D nearestSeg;
    protected double viewWidth;
    protected double scale;
    protected double fractionAlongPath;

    DPath path;

    public LinkSliderCalc(final DPath p, final double viewWidth) {
        this(p, 0.5, viewWidth);
    }

    /**
     * @param p 		path to slide along
     * @param flatness	optional - min distance between curve control points and flattend polyline representation
     * @param viewWidth width of camera viewport at scale == 1
     */
    public LinkSliderCalc(final DPath p, final double flatness, final double viewWidth) {
        this.path = p;
        this.polyLine = new PolyLine(this.path.getJava2DGeneralPath().getPathIterator(new AffineTransform(), flatness));
        this.viewWidth = viewWidth;
    }

    public DPath getPath() {
        return this.path;
    }

    /**
     * Call first to set the new mouse position.
     * This calculates the new position and scale factor along the path.
     * @param mousePos
     */
    public void updateMousePosition(final Point2D mousePos) {
        this.mousePos = new Point2D.Double(mousePos.getX(), mousePos.getX());

        // Nearest Point and tangent segment
        this.nearestSeg = this.polyLine.nearestSegment(mousePos);
        this.nearestPoint = PolyLine.nearestPointOnSegment(mousePos, this.nearestSeg);

        // Scale
        final double partialDist = this.polyLine.cumLengthAt(this.nearestPoint);
        final double pathLen = this.polyLine.getLength();

        // connes function
        // in 0..1 unit sqare: y = (1-(x*2-1)^6)^4
        final double x = partialDist / pathLen;
        final double y = Math.pow(1 - Math.pow(x * 2 - 1, 6), 4) * (1 - this.viewWidth / pathLen);
        this.scale = Math.min(1, Math.max(0.01, 1 - y)); // clamp .01 to 1

        // fraction along path
        this.fractionAlongPath = x;
    }

    /**
     * @return Closest point on curve to mouse position
     * (call updateMousePosition first)
     */
    public Point2D getPositionAlongPath() {
        return this.nearestPoint;
    }

    /**
     * @return fraction [0..1] closest point is along path
     * (call updateMousePosition first)
     */
    public double getFractionAlongPath() {
        return this.fractionAlongPath;
    }

    /**
     * @return scale [0.01 to 1] so that viewWidth == path length at midpoint. 
     * (call updateMousePosition first)
     */
    public double getScale() {
        return this.scale;
    }

    /**
     * @return tangent segment to curve at position along path 
     * (call updateMousePosition first)
     */
    public Line2D getTangentSegmentAtPathPosition() {
        return this.nearestSeg;
    }

}

class Vector2D {

    public double x;
    public double y;

    public Vector2D(final Point start, final Point end) {
        this.x = end.x - start.x;
        this.y = end.y - start.y;
    }

    public Vector2D(final Point2D start, final Point2D end) {
        this.x = end.getX() - start.getX();
        this.y = end.getY() - start.getY();
    }

    public Vector2D(final Vector2D start, final Vector2D end) {
        this.x = end.x - start.x;
        this.y = end.y - start.y;
    }

    public Vector2D(final Point p) {
        this.x = p.x;
        this.y = p.y;
    }

    public Vector2D() {
    }

    public Vector2D(final double _x, final double _y) {
        this.x = _x;
        this.y = _y;
    }

    public Vector2D(final Vector2D v) {
        this.x = v.x;
        this.y = v.y;
    }

    @Override
    public String toString() {
        return "(" + (int) this.x + "," + (int) this.y + ")";
    }

    public Vector2D duplicate() {
        return new Vector2D(this.x, this.y);
    }

    public static Vector2D normalize(final Vector2D v) {
        final double l = v.length();
        return new Vector2D(v.x / l, v.y / l);
    }

    public void normalize() {
        final double l = length();
        this.x /= l;
        this.y /= l;
    }

    public void negate() {
        this.x *= -1;
        this.y *= -1;
    }

    public static Vector2D negate(final Vector2D v) {
        return new Vector2D(-v.x, -v.y);
    }

    public double length() {
        return Math.sqrt(this.x * this.x + this.y * this.y);
    }

    static public Vector2D add(final Vector2D u, final Vector2D v) {
        return new Vector2D(u.x + v.x, u.y + v.y);
    }

    static public Vector2D add(final Vector2D u, final Vector2D v, final Vector2D w) {
        return new Vector2D(u.x + v.x + w.x, u.y + v.y + w.y);
    }

    public void add(final Vector2D v) {
        this.x += v.x;
        this.y += v.y;
    }

    static public Vector2D subtract(final Point u, final Vector2D v) {
        return new Vector2D(u.x - v.x, u.y - v.y);
    }

    static public Vector2D subtract(final Vector2D u, final Vector2D v) {
        return new Vector2D(u.x - v.x, u.y - v.y);
    }

    public void subtract(final Vector2D v) {
        this.x -= v.x;
        this.y -= v.y;
    }

    static public Vector2D multiply(final Vector2D v, final double m) {
        return new Vector2D(v.x * m, v.y * m);
    }

    public void multiply(final double m) {
        this.x *= m;
        this.y *= m;
    }

    static public double dot_product(final Vector2D u, final Vector2D v) {
        return u.x * v.x + u.y * v.y;
    }

    static public double cross_product(final Vector2D u, final Vector2D v) {
        return u.x * v.y - u.y * v.x;
    }

    static public double cos(final Vector2D u, final Vector2D v) {
        final double length = u.length() * v.length();

        if (length > 0) {
            return dot_product(u, v) / length;
        } else {
            return 0;
        }
    }

    static public double sin(final Vector2D u, final Vector2D v) {
        final double length = u.length() * v.length();

        if (length > 0) {
            return (cross_product(u, v)) / length;
        } else {
            return 0;
        }
    }

    // angle 0 to PI
    public static double get_angle0PI(final Vector2D u, final Vector2D v) {
        final double cosine = cos(u, v);
        if (cosine <= -1) {
            return Math.PI;
        } else if (cosine >= 1) {
            return 0;
        } else {
            return Math.acos(cosine);
        }
    }

    // angle -PI to PI
    public double get_anglePIPI(final Vector2D v) {
        final double cos = cos(this, v);
        final double sin = sin(this, v);
        if (cos == 0) {
            if (sin > 0) {
                return Math.PI / 2;
            } else {
                return -Math.PI / 2;
            }
        }
        if (sin == 0) {
            if (cos > 0) {
                return 0;
            } else {
                return Math.PI;
            }
        }

        // -PI/2   ...   +PI/2
        double angle = Math.atan(sin / cos);

        if (cos < 0) {
            if (angle > 0) {
                angle -= Math.PI;
            } else {
                angle += Math.PI;
            }
        }
        return angle;
    }

    public static Vector2D rotate90(final Vector2D vec) {
        return new Vector2D(vec.y, -vec.x);
    }

    public static Vector2D rotate(final Vector2D v, final double radians) {

        final double cos = Math.cos(radians);
        final double sin = Math.sin(radians);

        return new Vector2D(v.x * cos - v.y * sin, v.x * sin + v.y * cos);
    }

    public static Vector2D flip(final Vector2D vec, final Vector2D axis) {
        final Vector2D x_vec = Vector2D.normalize(axis);

        final Vector2D y_vec = Vector2D.rotate90(x_vec);
        final double x = Vector2D.dot_product(vec, x_vec);
        double y = Vector2D.dot_product(vec, y_vec);
        y = -y;

        return Vector2D.add(Vector2D.multiply(x_vec, x), Vector2D.multiply(y_vec, y));
    }

    public static double distance(final int x1, final int y1, final int x2, final int y2) {
        return Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
    }

    public static double distance(final Point start, final Point end) {
        return distance(start.x, start.y, end.x, end.y);
    }

    public static double distance(final double x1, final double y1, final double x2, final double y2) {
        return Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
    }

    public Point point() {
        return new Point((int) this.x, (int) this.y);
    }

    static public Point translatePoint(final Point p, final Vector2D vec) {
        return new Point((int) (p.x + vec.x), (int) (p.y + vec.y));
    }
}

class PolyLine {

    ArrayList<Line2D> segments; // polyline segments
    ArrayList<Double> cumLengths; // total length until seg i (exclusive)
    double length;

    public PolyLine(final PathIterator pi) {
        // get segments
        this.segments = new ArrayList<Line2D>();
        final double[] coords = {0, 0, 0, 0, 0, 0 };
        Point2D lastPt = new Point2D.Double();
        while (!pi.isDone()) {
            switch (pi.currentSegment(coords)) {
            case PathIterator.SEG_MOVETO:
                lastPt = new Point2D.Double(coords[0], coords[1]);
                break;
            case PathIterator.SEG_LINETO:
                final Point2D.Double newPt = new Point2D.Double(coords[0], coords[1]);
                this.segments.add(new Line2D.Double(lastPt, newPt));
                lastPt = newPt;
                break;
            }
            pi.next();
        }
        // precalc lengths
        calcCumLengths();
    }

    // calculate and store length of each segment
    public void calcCumLengths() {
        this.length = 0;
        this.cumLengths = new ArrayList<Double>(this.segments.size());
        for (int i = 0; i < this.segments.size(); ++i) {
            this.cumLengths.add(new Double(this.length));
            this.length += segmentLength(this.segments.get(i));
        }
    }

    // cumulative length -- assumes point is on path
    public double cumLengthAt(final Point2D pt) {
        final int segi = nearestSegmentIdx(pt);
        //Point2D nearestPt = nearestPointOnSegment(pt, segments.get(segi));
        final Point2D segPt = this.segments.get(segi).getP1();
        final double dx = pt.getX() - segPt.getX();
        final double dy = pt.getY() - segPt.getY();
        return this.cumLengths.get(segi) + Math.sqrt(dx * dx + dy * dy);
    }

    public double getLength() {
        return this.length;
    }

    public double getLengthAtSegmentIdx(final int idx) {
        return this.cumLengths.get(idx);
    }

    public Point2D nearestPoint(final Point2D pt) {
        return nearestPointOnSegment(pt, nearestSegment(pt));
    }

    public Line2D nearestSegment(final Point2D pt) {
        return this.segments.get(nearestSegmentIdx(pt));
    }

    public int nearestSegmentIdx(final Point2D pt) {
        double shortestDist = java.lang.Double.POSITIVE_INFINITY;
        int nearestSegIdx = 0;
        int i = 0;
        for (i = 0; i < this.segments.size(); ++i) {
            final double dist = this.segments.get(i).ptSegDist(pt);
            if (dist < shortestDist) {
                shortestDist = dist;
                nearestSegIdx = i;
            }
        }
        return nearestSegIdx;
    }

    // Calculate closest point on segment
    public static Point2D nearestPointOnSegment(final Point2D pt, final Line2D seg) {
        final Point2D p1 = seg.getP1();
        final Point2D p2 = seg.getP2();
        final Vector2D segv = new Vector2D(p1, p2);
        final double len = segv.length();
        segv.normalize();
        final Vector2D ptv = new Vector2D(p1, pt);
        final double t = Vector2D.dot_product(ptv, segv) / len;
        Point2D closestPt = p1;
        if (t > 1) {
            closestPt = p2;
        } else if (t > 0) {
            closestPt = new Point2D.Double((1 - t) * p1.getX() + t * p2.getX(), (1 - t) * p1.getY() + t * p2.getY());
        }
        return closestPt;

    }

    /*
    public ArrayList<Point2D> getIntersections(PolyLine poly){
    	ArrayList<Point2D> intersections = new ArrayList<Point2D>();
    
    	for(int i = 0; i < this.segments.size(); ++i){
    		Line2D seg1 = this.segments.get(i);
    		for(int j = 0; j < poly.segments.size(); ++j){
    			Line2D seg2 = poly.segments.get(j);
    			if(seg1.intersectsLine(poly.segments.get(j))){
    				//TODO intersects?  intersections.add(seg1.);
    			}
    				
    		}
    				
    	}
    	
    	return intersections;
    }*/
    public static Point2D SegmentSegmentIntersection(final Line2D l1, final Line2D l2) {
        final double x1 = l1.getX1();
        final double y1 = l1.getY1();
        final double x2 = l1.getX1();
        final double y2 = l1.getY2();
        final double xx1 = l2.getX1();
        final double yy1 = l2.getY1();
        final double xx2 = l2.getX2();
        final double yy2 = l2.getY2();

        // a * x + b * y + c = 0
        double a0, b0, c0, a1, b1, c1;
        a0 = y1 - y2;
        b0 = x2 - x1;
        c0 = y2 * x1 - x2 * y1;
        a1 = yy1 - yy2;
        b1 = xx2 - xx1;
        c1 = yy2 * xx1 - xx2 * yy1;

        if (Math.abs(a0 * b1 - a1 * b0) < 0.00001) {
            return null;
        }

        final double x = ((b0 * c1 - b1 * c0) / (a0 * b1 - a1 * b0));
        final double y = ((a0 * c1 - a1 * c0) / (a1 * b0 - a0 * b1));
        return new Point2D.Double(x, y);

    }

    // Calculate the intersection of a ray and a circle
    // Segment seg, circle center at cc, radius r
    // Zero to two intersections are returned  in int1, int 2
    // Return FALSE if no intersection
    public boolean rayCircleIntersections(
        final Line2D seg,
        final Point2D cc,
        final double r,
        final Point2D int1,
        final Point2D int2)
    {
        final Point2D p1 = seg.getP1();
        final Point2D p2 = seg.getP2();
        double mu1, mu2;
        double a, b, c;
        double bb4ac;
        Point2D.Double dp;

        dp = new Point2D.Double(p2.getX() - p1.getX(), p2.getY() - p1.getY());

        a = dp.getX() * dp.getX() + dp.getY() * dp.getY();
        b = 2 * (dp.getX() * (p1.getX() - cc.getX()) + dp.getY() * (p1.getY() - cc.getY()));
        c = cc.getX() * cc.getX() + cc.getY() * cc.getY();
        c += p1.getX() * p1.getX() + p1.getY() * p1.getY();
        c -= 2 * (cc.getX() * p1.getX() + cc.getY() * p1.getY());
        c -= r * r;
        bb4ac = b * b - 4 * a * c;
        if (Math.abs(a) < 0.00001 || bb4ac < 0) {
            int1.setLocation(Double.NaN, Double.NaN);
            int2.setLocation(Double.NaN, Double.NaN);
            return false;
        }

        mu1 = (-b + Math.sqrt(bb4ac)) / (2 * a);
        mu2 = (-b - Math.sqrt(bb4ac)) / (2 * a);
        final Vector2D v1 = new Vector2D(p1, p2);
        final Vector2D v2 = new Vector2D(p1, p2);
        v1.multiply(mu1);
        v2.multiply(mu2);
        int1.setLocation(p1.getX() + v1.x, p1.getY() + v1.y);
        int2.setLocation(p2.getX() + v2.x, p2.getY() + v1.y);
        return true;
    }

    // finds an intersection with a circle of radius r, centered on point 0.
    // if found, returns index of intersecting segment, and places intersection pt in intPt
    // if no intersection is found, a new radius is calculated: altR * distance between endpoints
    // if something is wrong, returns -1;
    public int intersectsCircleOnStartPoint(final double r, final double altR, final Point2D intPt) {
        final Point2D center = this.segments.get(0).getP1();
        final Point2D pt1 = new Point2D.Double();
        final Point2D pt2 = new Point2D.Double();
        for (int i = 0; i < this.segments.size(); ++i) {
            if (center.distance(this.segments.get(i).getP2()) < r) {
                continue;
            }
            if (rayCircleIntersections(this.segments.get(i), center, r, pt1, pt2)) {
                intPt.setLocation(pt1);
                return i;
            }
        }
        // No intersection, calc new r
        final Double shortR = altR * center.distance(this.segments.get(this.segments.size() - 1).getP2());
        for (int i = 0; i < this.segments.size(); ++i) {
            if (center.distance(this.segments.get(i).getP2()) < shortR) {
                continue;
            }
            if (rayCircleIntersections(this.segments.get(i), center, shortR, pt1, pt2)) {
                intPt.setLocation(pt1);
                return i;
            }
        }
        return -1;

    }

    public static final double segmentLength(final Line2D seg) {
        final double dx = seg.getX2() - seg.getX1();
        final double dy = seg.getY2() - seg.getY1();
        return Math.sqrt(dx * dx + dy * dy);
    }

    public Point2D firstPoint() {
        return this.segments.get(0).getP1();
    }

    public Point2D lastPoint() {
        return this.segments.get(this.segments.size() - 1).getP2();
    }
}
