/**
 * Displays an export dialogue similar to photoshop
 * 
 * @author shu ning bian
 * @version 07.02.2005
 *
 * Licensed under the GNU LGPL. For full terms see the file COPYING.
 *
 * $Id$
 */

package net.claribole.zgrviewer;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;

import javax.imageio.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;

public class PNGExportWindow {

    private final static String units[] = {"pixels", "inches", "cm", "mm" };
    private final static int PIXEL = 0;
    private final static int INCH = 1;
    private final static int CM = 2;
    private final static int MM = 3;

    private final static double CM_PER_INCH = 2.54;
    private final static double MM_PER_INCH = 25.4;

    private int last_dpi = 72;
    private int last_width_unit = PIXEL;
    private int last_height_unit = PIXEL;

    private JFrame frame;

    private JPanel content;
    private JPanel options;
    private JPanel fileSelection;
    private JPanel buttons;

    private JButton export;
    private JButton cancel;
    private JButton browse;

    private TitledBorder optionsBorder;

    private JTextField filePath;
    private SpinnerNumberModel width;
    private SpinnerNumberModel height;

    private JComboBox widthUnit;
    private JComboBox heightUnit;

    private SpinnerNumberModel dpiSpinner;

    private FlowLayout layout;

    private double realWidth;
    private double realHeight;

    private static File selectedFile = null;
    /* since using setValue causes a stateChange event, we need to 
     * stop recursion by locking the state so subsequent functions
     * called by the changeStated handlers quietly exists
     */

    boolean stateLock = false;

    private int dpi = this.last_dpi;

    GraphicsManager grMngr;

    /**
     * Constructor for objects of class PNGExportWindow
     * @param   w   starting pixel width
     * @param   h   starting pixel height
     *@param gm GraphicsManager instantiated by the parent ZGRViewer/ZGRApplet
     */
    public PNGExportWindow(final double w, final double h, final GraphicsManager gm) {
        this.grMngr = gm;
        if (w < 1 || h < 1) {
            JOptionPane.showMessageDialog(
                this.grMngr.mainView.getFrame(),
                "Can not export visible region of size 0.",
                "Export to PNG error",
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        this.realWidth = w;
        this.realHeight = h;

        // generate the window, but don't show it
        this.frame = new JFrame("PNG Export Options");
        this.frame.setResizable(false);

        // construct the content pane
        this.content = new JPanel(new BorderLayout());
        this.frame.setContentPane(this.content);

        // construct the file selection pane
        this.fileSelection = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        this.filePath = new JTextField();
        this.filePath.setColumns(19);

        if (selectedFile != null) {
            this.filePath.setText(selectedFile.getPath());
        }

        class BrowseActionListener implements ActionListener {
            @Override
            public void actionPerformed(final ActionEvent e) {
                JFileChooser fc;

                if (selectedFile == null) {
                    fc = new JFileChooser();
                } else {
                    fc = new JFileChooser(selectedFile.getParentFile());
                }

                fc.setDialogTitle("Export PNG");
                if (fc.showSaveDialog(PNGExportWindow.this.frame) == JFileChooser.APPROVE_OPTION) {
                    // set the filePath text box
                    selectedFile = fc.getSelectedFile();

                    if (selectedFile != null) {
                        PNGExportWindow.this.filePath.setText(selectedFile.getPath());
                        PNGExportWindow.this.export.setEnabled(true);
                    } else {
                        PNGExportWindow.this.export.setEnabled(false);
                        PNGExportWindow.this.filePath.setText("");
                    }
                }
            }
        }
        this.browse = new JButton("Browse");
        this.browse.addActionListener(new BrowseActionListener());

        this.fileSelection.add(new JLabel("File path: ", SwingConstants.RIGHT));
        this.fileSelection.add(this.filePath);
        this.fileSelection.add(this.browse);

        this.content.add(this.fileSelection, BorderLayout.NORTH);

        // create the options panel
        this.options = new JPanel(new GridLayout(3, 3));
        this.optionsBorder = BorderFactory.createTitledBorder("Export options");
        this.options.setBorder(this.optionsBorder);

        // define some helper classes
        class UnitActionListener implements ActionListener {
            @Override
            public void actionPerformed(final ActionEvent e) {
                updateTextFields(false);
            }
        }

        class DPIChangeListener implements ChangeListener {
            @Override
            public void stateChanged(final ChangeEvent e) {
                PNGExportWindow.this.dpi = PNGExportWindow.this.dpiSpinner.getNumber().intValue();
                updateTextFields(false);
                PNGExportWindow.this.last_dpi = PNGExportWindow.this.dpi;
            }
        }

        class TextChangeListener implements ChangeListener {
            @Override
            public void stateChanged(final ChangeEvent e) {
                updateTextFields(true);
            }
        }

        final UnitActionListener unitListener = new UnitActionListener();
        final TextChangeListener textListener = new TextChangeListener();

        this.width = new SpinnerNumberModel(this.realWidth, 1, Double.MAX_VALUE, 1.0);
        this.width.addChangeListener(textListener);
        this.widthUnit = new JComboBox(units);
        this.widthUnit.addActionListener(unitListener);

        this.options.add(new JLabel("Image width: ", SwingConstants.RIGHT));
        this.options.add(new JSpinner(this.width));
        this.options.add(this.widthUnit);

        this.height = new SpinnerNumberModel(this.realHeight, 1, Double.MAX_VALUE, 1.0);
        this.height.addChangeListener(textListener);
        this.heightUnit = new JComboBox(units);
        this.heightUnit.addActionListener(unitListener);

        this.options.add(new JLabel("Image height: ", SwingConstants.RIGHT));
        this.options.add(new JSpinner(this.height));
        this.options.add(this.heightUnit);

        this.dpiSpinner = new SpinnerNumberModel(this.last_dpi, 72, 9999, 1);
        this.dpiSpinner.addChangeListener(new DPIChangeListener());

        this.options.add(new JLabel("Dots per inch: ", SwingConstants.RIGHT));
        this.options.add(new JSpinner(this.dpiSpinner));

        this.content.add(this.options, BorderLayout.CENTER);

        // create the buttons panel
        this.buttons = new JPanel(new FlowLayout(FlowLayout.RIGHT));

        class CancelActionListener implements ActionListener {
            @Override
            public void actionPerformed(final ActionEvent e) {
                hide();
            }
        }

        class ExportActionListener implements ActionListener {
            @Override
            public void actionPerformed(final ActionEvent e) {
                hide();
                PNGExportWindow.this.grMngr.mainView.setStatusBarText("Exporting to PNG "
                    + PNGExportWindow.this.filePath.getText()
                    + " ... (This operation can take some time)");

                PNGExportWindow.this.realWidth =
                    unitToPixel(PNGExportWindow.this.realWidth, PNGExportWindow.this.widthUnit.getSelectedIndex());
                PNGExportWindow.this.realHeight =
                    unitToPixel(PNGExportWindow.this.realHeight, PNGExportWindow.this.heightUnit.getSelectedIndex());

                final Vector layers = new Vector();
                layers.add(PNGExportWindow.this.grMngr.mainCamera);
                final BufferedImage bi =
                    PNGExportWindow.this.grMngr.mainView.rasterize(
                        (int) PNGExportWindow.this.realWidth,
                        (int) PNGExportWindow.this.realHeight,
                        layers);
                try {
                    ImageIO.write(bi, "png", new File(PNGExportWindow.this.filePath.getText()));
                } catch (final IOException ex) {
                    System.err.println("ZGRViewer::Error while exporting current view to PNG file "
                        + PNGExportWindow.this.filePath.getText());
                    ex.printStackTrace();
                }
                PNGExportWindow.this.grMngr.mainView.setStatusBarText("Exporting to PNG "
                    + PNGExportWindow.this.filePath.getText()
                    + " ...done");
            }
        }

        this.export = new JButton("Export");
        this.export.addActionListener(new ExportActionListener());
        this.export.setEnabled(false);

        this.cancel = new JButton("Cancel");
        this.cancel.addActionListener(new CancelActionListener());

        this.buttons.add(this.cancel);
        this.buttons.add(this.export);

        this.content.add(this.buttons, BorderLayout.SOUTH);

        updateTextFields(false);

        // prepare it for display
        this.frame.getRootPane().setDefaultButton(this.export);
        this.content.setPreferredSize(new Dimension(400, 200));
        this.frame.pack();
        this.frame.setLocationRelativeTo(this.grMngr.mainView.getFrame());
        this.frame.setVisible(true);
    }

    /**
     * Displays the window 
     */
    public void show() {
        this.frame.setVisible(true);
    }

    /** 
     * Hides the window
     */
    public void hide() {
        this.frame.setVisible(false);
    }

    /**
     * updates all the text fields according to new selected unit and dpi informaion
     * @param   textChange  whether or not the text fields have changed
     */
    private void updateTextFields(final boolean textChange) {
        // only one instance of this method, ever
        if (this.stateLock) {
            return;
        }

        this.stateLock = true;

        String text = "";

        if (textChange) {
            this.realWidth = this.width.getNumber().doubleValue();
            this.realHeight = this.height.getNumber().doubleValue();
        }

        // calculate what to display for width
        double pixels = this.realWidth;

        // found out how many pixels we had before
        pixels = unitToPixel(this.realWidth, this.last_width_unit);
        pixels = pixelToUnit(pixels, this.widthUnit.getSelectedIndex());

        if (this.widthUnit.getSelectedIndex() == PIXEL) {
            text = "" + (int) pixels;
        } else {
            text = "" + pixels;
        }

        if (text.indexOf(".") >= 0) {
            text = text.substring(0, text.indexOf(".") + 2);
        }
        this.width.setValue(new Double(text));

        // record the real value to avoid loss of precision during unit conversion
        this.realWidth = pixels;

        // now do height
        pixels = this.realHeight;

        // found out how many pixels we had before
        pixels = unitToPixel(this.realHeight, this.last_height_unit);
        pixels = pixelToUnit(pixels, this.heightUnit.getSelectedIndex());

        if (this.heightUnit.getSelectedIndex() == PIXEL) {
            text = "" + (int) pixels;
        } else {
            text = "" + pixels;
        }

        if (text.indexOf(".") >= 0) {
            text = text.substring(0, text.indexOf(".") + 2);
        }
        this.height.setValue(new Double(text));

        // record the real value to avoid loss of precision during unit conversion
        this.realHeight = pixels;

        // set the last unit information
        this.last_width_unit = this.widthUnit.getSelectedIndex();
        this.last_height_unit = this.heightUnit.getSelectedIndex();

        this.stateLock = false;
    }

    /** Converts the given unit into pixels
     *	@param	number	number to convert
     *	@param	unit	unit to convert from
     *	@return			number of pixels 
     */
    private double unitToPixel(double number, final int unit) {
        //System.out.print("converting " + number + " " + units[unit] + " to " );
        switch (unit) {
        case INCH:
            number *= this.last_dpi;
            break;

        case PIXEL:
            break;

        case CM:
            number = number / CM_PER_INCH * this.last_dpi;
            break;

        case MM:
            number = number / MM_PER_INCH * this.last_dpi;
            break;
        }
        //System.out.println(number + " pixels");
        return number;
    }

    /** Converts pixels to the given unit
     *	@param	pixels	number of pixels to convert
     *	@param	unit	unit to conver to
     *	@return			number of units the given number of pixels represent
     */
    private double pixelToUnit(double pixels, final int unit) {
        //System.out.print("converting " + pixels + " pixels to ");
        switch (unit) {
        case INCH:
            pixels = pixels / this.last_dpi;
            break;

        case PIXEL:
            pixels = pixels * this.dpi / this.last_dpi;
            break;

        case CM:
            pixels = pixels / this.last_dpi * CM_PER_INCH;
            break;

        case MM:
            pixels = pixels / this.last_dpi * MM_PER_INCH;
            break;
        }
        //System.out.println(pixels +" " + units[unit]);
        return pixels;
    }
}
