/*   FILE: TextViewer.java
 *   DATE OF CREATION:   Wed Dec 03 09:31:46 2003
 *   Copyright (c) Emmanuel Pietriga, 2002. All Rights Reserved
 *   Copyright (c) INRIA, 2004-2011. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 *
 * $Id$
 */

package net.claribole.zgrviewer;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

/*A simple text viewer that displays the content of a stringbuffer. Can be set to automatically update its content periodically. Can be used for showing error logs, raw source files,...*/

public class TextViewer extends JFrame implements ActionListener, KeyListener, Runnable {

    Thread runView;

    JButton okBt;
    JButton clearBt;

    StringBuffer text;
    int oldSize;

    int period;

    JTextArea ar;

    /**
     *@param msgs text displayed in main area
     *@param frameTitle string appearing in window title bar
     *@param d periodic update (in milliseconds) of the main area w.r.t msgs (0=no update)
     *@param clear show a clear button in addition to the OK button
     */
    public TextViewer(final StringBuffer msgs, final String frameTitle, final int d, final boolean clear) {
        this.text = msgs;
        this.oldSize = this.text.length();
        this.period = d;
        this.ar = new JTextArea(this.text.toString());
        this.ar.setFont(ConfigManager.defaultFont);
        this.ar.setEditable(false);
        this.ar.setLineWrap(true);
        final JScrollPane sp = new JScrollPane(this.ar);
        sp.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        //sp.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        final Container cpane = this.getContentPane();
        final GridBagLayout gridBag = new GridBagLayout();
        final GridBagConstraints constraints = new GridBagConstraints();
        constraints.fill = GridBagConstraints.BOTH;
        constraints.anchor = GridBagConstraints.CENTER;
        cpane.setLayout(gridBag);
        buildConstraints(constraints, 0, 0, 2, 1, 100, 98);
        gridBag.setConstraints(sp, constraints);
        cpane.add(sp);
        if (clear) {
            this.okBt = new JButton("OK");
            this.okBt.addActionListener(this);
            this.okBt.addKeyListener(this);
            constraints.anchor = GridBagConstraints.EAST;
            constraints.fill = GridBagConstraints.NONE;
            buildConstraints(constraints, 0, 1, 1, 1, 50, 2);
            gridBag.setConstraints(this.okBt, constraints);
            cpane.add(this.okBt);
            this.clearBt = new JButton("Clear");
            this.clearBt.addActionListener(this);
            this.clearBt.addKeyListener(this);
            constraints.anchor = GridBagConstraints.WEST;
            buildConstraints(constraints, 1, 1, 1, 1, 50, 0);
            gridBag.setConstraints(this.clearBt, constraints);
            cpane.add(this.clearBt);
        } else {
            this.okBt = new JButton("OK");
            this.okBt.addActionListener(this);
            this.okBt.addKeyListener(this);
            constraints.anchor = GridBagConstraints.CENTER;
            constraints.fill = GridBagConstraints.HORIZONTAL;
            buildConstraints(constraints, 0, 1, 2, 1, 100, 2);
            gridBag.setConstraints(this.okBt, constraints);
            cpane.add(this.okBt);
        }
        final WindowListener w0 = new WindowAdapter() {
            @Override
            public void windowClosing(final WindowEvent e) {
                stop();
                dispose();
            }
        };
        this.addWindowListener(w0);
        this.setTitle(frameTitle);
        this.pack();
        this.setLocation(0, 0);
        this.setSize(700, 300);
        this.setVisible(true);
        this.okBt.requestFocus();
        if (this.period > 0) {
            addHierarchyListener(new HierarchyListener() {
                @Override
                public void hierarchyChanged(final HierarchyEvent e) {
                    if (isShowing()) {
                        start();
                    } else {
                        stop();
                    }
                }
            });
            start();
        }
    }

    /**
     *@param msgs text displayed in main area
     *@param frameTitle string appearing in window title bar
     *@param d periodic update (in milliseconds) of the main area w.r.t msgs (0=no update)
     *@param x position of top-left corner
     *@param y position of top-left corner
     *@param width frame width
     *@param height frame height
     *@param clear show a clear button in addition to the OK button
     */
    public TextViewer(
        final StringBuffer msgs,
        final String frameTitle,
        final int d,
        final int x,
        final int y,
        final int width,
        final int height,
        final boolean clear)
    {
        this.text = msgs;
        this.oldSize = this.text.length();
        this.period = d;
        this.ar = new JTextArea(this.text.toString());
        this.ar.setFont(ConfigManager.defaultFont);
        this.ar.setEditable(false);
        this.ar.setLineWrap(true);
        this.ar.setWrapStyleWord(true);
        final JScrollPane sp = new JScrollPane(this.ar);
        sp.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        //sp.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        final Container cpane = this.getContentPane();
        final GridBagLayout gridBag = new GridBagLayout();
        final GridBagConstraints constraints = new GridBagConstraints();
        constraints.fill = GridBagConstraints.BOTH;
        constraints.anchor = GridBagConstraints.CENTER;
        cpane.setLayout(gridBag);
        buildConstraints(constraints, 0, 0, 2, 1, 100, 98);
        gridBag.setConstraints(sp, constraints);
        cpane.add(sp);
        if (clear) {
            this.okBt = new JButton("OK");
            this.okBt.addActionListener(this);
            this.okBt.addKeyListener(this);
            constraints.anchor = GridBagConstraints.EAST;
            constraints.fill = GridBagConstraints.NONE;
            buildConstraints(constraints, 0, 1, 1, 1, 50, 2);
            gridBag.setConstraints(this.okBt, constraints);
            cpane.add(this.okBt);
            this.clearBt = new JButton("Clear");
            this.clearBt.addActionListener(this);
            this.clearBt.addKeyListener(this);
            constraints.anchor = GridBagConstraints.WEST;
            buildConstraints(constraints, 1, 1, 1, 1, 50, 0);
            gridBag.setConstraints(this.clearBt, constraints);
            cpane.add(this.clearBt);
        } else {
            this.okBt = new JButton("OK");
            this.okBt.addActionListener(this);
            this.okBt.addKeyListener(this);
            constraints.anchor = GridBagConstraints.CENTER;
            constraints.fill = GridBagConstraints.HORIZONTAL;
            buildConstraints(constraints, 0, 1, 2, 1, 100, 2);
            gridBag.setConstraints(this.okBt, constraints);
            cpane.add(this.okBt);
        }
        final WindowListener w0 = new WindowAdapter() {
            @Override
            public void windowClosing(final WindowEvent e) {
                stop();
                dispose();
            }
        };
        this.addWindowListener(w0);
        this.setTitle(frameTitle);
        this.pack();
        this.setLocation(x, y);
        this.setSize(width, height);
        this.setVisible(true);
        this.okBt.requestFocus();
        if (this.period > 0) {
            addHierarchyListener(new HierarchyListener() {
                @Override
                public void hierarchyChanged(final HierarchyEvent e) {
                    if (isShowing()) {
                        start();
                    } else {
                        stop();
                    }
                }
            });
            start();
        }
    }

    public void start() {
        this.runView = new Thread(this);
        this.runView.setPriority(Thread.MIN_PRIORITY);
        this.runView.start();
    }

    public synchronized void stop() {
        this.runView = null;
        notify();
    }

    @Override
    public void run() {
        final Thread me = Thread.currentThread();
        while (this.runView == me) {
            if (this.text.length() != this.oldSize) {
                this.ar.setText(this.text.toString());
            } //update text only if new error messages
            this.oldSize = this.text.length();
            try {
                Thread.sleep(this.period);
            } catch (final InterruptedException e) {
            }
        }
    }

    @Override
    public void actionPerformed(final ActionEvent e) {
        if (e.getSource() == this.okBt) {
            this.stop();
            this.dispose();
        } else if (e.getSource() == this.clearBt) {
            this.text.setLength(0);
            this.ar.setText("");
        }
    }

    @Override
    public void keyPressed(final KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_ENTER) {
            if (e.getSource() == this.okBt) {
                this.stop();
                this.dispose();
            } else if (e.getSource() == this.clearBt) {
                this.text.setLength(0);
                this.ar.setText("");
            }
        }
    }

    @Override
    public void keyReleased(final KeyEvent e) {
    }

    @Override
    public void keyTyped(final KeyEvent e) {
    }

    void buildConstraints(
        final GridBagConstraints gbc,
        final int gx,
        final int gy,
        final int gw,
        final int gh,
        final int wx,
        final int wy)
    {
        gbc.gridx = gx;
        gbc.gridy = gy;
        gbc.gridwidth = gw;
        gbc.gridheight = gh;
        gbc.weightx = wx;
        gbc.weighty = wy;
    }

}
